/*************************************************************************
 *************************************************************************
 **									**
 **	 		Copyright (c) 1994, 1995 			**
 **	    Brown University Institute for Medical Computing		**
 **	      and the American College of Cardiology (ACC)		**
 **									**
 ** The software and support documentation for the American College of	**
 ** Cardiology's 1995 demonstration of DICOM interchange media were	**
 ** developed at the Brown University Institute for Medical Computing.	**
 **									** 
 ** THIS SOFTWARE IS MADE AVAILABLE "AS IS", AND NEITHER THE AMERICAN	**
 ** COLLEGE OF CARDIOLOGY (ACC) NOR BROWN UNIVERSITY MAKE ANY WARRANTY	**
 ** ABOUT THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR FITNESS	**
 ** FOR ANY PARTICULAR USE, FREEDOM FROM DEFECTS, OR ITS CONFORMANCE TO	**
 ** ANY SPECIFICATION.  THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE	**
 ** OF THE SOFTWARE IS WITH THE USER.					**
 **									** 
 ** Copyright of the software and supporting documentation is jointly	**
 ** owned by the ACC and Brown University, and free access is hereby	**
 ** granted as a license to use this software, copy this software and	**
 ** prepare derivative works based upon this software.  However, any	**
 ** distribution of the source code or executable versions of this	**
 ** software, or of the	 supporting documentation (or of derivative	**
 ** works of the software or supporting documentation) must include	**
 ** this copyright notice and explanation.				**
 **									** 
 ** Author: Jonathan L. Elion MD, FACC					**
 **									**
 ************************************************************************* 
 *************************************************************************/ 

/************************************************************************
 *									*
 * This utility reads a script file and uses the information contained	*
 * therein to create DICOM files and the corresponding DICOMDIR.  As	*
 * such, this utility acts as a DICOM "File Set Creator" (the defined	*
 * role of a File Set Creator is to create the DICOMDIR File and zero	*
 * or more DICOM files).						*
 *									*
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "acc95.h"


/************************************************************************
 * The following items are GLOBAL, and are used by several routines	*
 * for reading and parsing the script files, and writing output files.	*
 ************************************************************************/

FILE	*scriptFile; 			/* Script input file */
FILE	*outFile;			/* Output file */
int	lineCount = 0, 			/* Number of lines read */
	scriptFileOpen = FALSE,		/* Is the input file open? */
	outFileOpen = FALSE,		/* Is the output file open? */
	thisEndian,			/* "Endian"ism of host CPU */
	readType,			/* PARSING or SCANNING script */
	explicit_VR = FALSE,		/* Using explicit VR mode? */
	encodedPixels = FALSE,		/* Will pixel data get encoded? */
	symbolCount;			/* # of symbols in symbol table */
char	scriptTag[MAX_LINE_LENGTH],	/* First valid word found on line */
	scriptData[MAX_LINE_LENGTH];	/* Second valid word found on line */
UnsignedLong byteCount;		/* # bytes if DICOM messages written */

struct {				/* Symbol table for addresses */
    char name[32];
    UnsignedLong value;
} symbolTable[MAX_SYMBOLS];

char *VR_label[] = {			/* Value Representations (printable) */
    "AE","AS","AT","CS","DA","DS","DT","FL","FD","IS","LO","LT","OB",
    "OW","PN","SH","SL","SQ","SS","ST","TM","UI","UL","US","**","??"};


/************************************************************************
 *									*
 * The badScript() routine is used to bail out from reading the script	*
 * file when an error is found.  An error message is printed, the	*
 * script file closed, and the program exits.  Simple.			*
 *									*
 ************************************************************************/

void badScript(char *msg)
{
    fprintf(stderr, "Error on line %d of script file: %s\n", lineCount, msg);
    if (scriptFileOpen == TRUE) fclose(scriptFile);
    if (outFileOpen == TRUE) fclose(outFile);
    exit(EXIT_FAILURE);
}


/************************************************************************
 *									*
 * The lookupSymbol() routine looks up the value of a symbol, and	*
 * returns its value (an UnsignedLong corresponding to the offset of	*
 * the label's offset from the beginning of the file.)		* 
 *									*
 ************************************************************************/

UnsignedLong lookupSymbol(char *value)
{
    int i;				/* for looping */
    char msg[80];			/* Used to construct error messages */

    /* Do a simple exhaustive search -- symbol table is NOT sorted */
    if (symbolCount > 0) for (i=0; i<symbolCount; i++) 
    {
	if (strcmp(symbolTable[i].name, value) == 0)
	    return(symbolTable[i].value); /* if names match, return value */
    }
    sprintf(msg, "Undefined symbol: %s", value);
    badScript(msg);			/* Not found -- error */
    return 0; /* Never reached, just prevents compiler complaint */
}


/************************************************************************
 *									*
 * The readScript() routine reads lines from the script file.  Blank	*
 * lines are skipped (including any line with only a comment.  When a	*
 * valid line is found, it is parsed.  The routine returns the number	*
 * words found (or 0 when End-of-file is found).			*
 *									*
 ************************************************************************/

int readScript(void)
{
    char *oneWord;
    char scriptLine[MAX_LINE_LENGTH];	/* Script file input lines */
    int i, len, wordCount;

    do {
	if (fgets(scriptLine, MAX_LINE_LENGTH, scriptFile) == NULL) return 0;
	lineCount++;
	wordCount = 0;
	oneWord = scriptLine;

	/* Skip leading white space, check for empty line */
	while ((oneWord[0] == ' ') || (oneWord[0] == '\t')) oneWord++;
	if ((oneWord[0] == '\n') || (oneWord[0] == ';')) continue;

	/* Find the first "word" on the line -- move it to scriptTag */
	len = strcspn(oneWord, " \t\n;"); /* Length of the word */
	strncpy(scriptTag, oneWord, len); /* Copy it */
	scriptTag[len] = '\0'; /* Terminate the word */
	for (i=0; i<len; i++) scriptTag[i] = toupper(scriptTag[i]);
	oneWord = strpbrk(oneWord, " \t\n;"); /* Restart scan after it */ 
	wordCount++;

	/* Skip white space between words 1 and 2, check for end of line */
	while ((oneWord[0] == ' ') || (oneWord[0] == '\t')) oneWord++;
	if ((oneWord[0] == '\n') || (oneWord[0] == ';')) continue;

	/* Handle a quoted string */
	if (oneWord[0] == '\"')
	{
	    oneWord++; /* skip of leading " mark */
	    if (strchr(oneWord, '\"') == NULL) badScript("Unmatched quote");
	    len = strcspn(oneWord, "\""); /* Length of the quote */
	    strncpy(scriptData, oneWord, len); /* Copy the quote */
	    scriptData[len] = '\0'; /* Terminate the the quote */
	    oneWord = strpbrk(oneWord, "\""); /* Restart scan after quote */ 
	    oneWord++;
	}
	else /* Handle non-quoted data */
	{
	    len = strcspn(oneWord, " \t\n;"); /* Length of the data */
	    strncpy(scriptData, oneWord, len); /* Copy it */
	    scriptData[len] = '\0'; /* Terminate the data */
	    for (i=0; i<len; i++) scriptData[i] = toupper(scriptData[i]);
	    oneWord = strpbrk(oneWord, " \t\n;"); /* Restart scan after it */ 
	}
	wordCount++;

	/* Make sure there's nothing left at the end of the line */
	while ((oneWord[0] == ' ') || (oneWord[0] == '\t')) oneWord++;
	if ((oneWord[0] == '\n') || (oneWord[0] == ';')) continue;
	if (readType == PARSING)
	    printf("Warning: Extra characters ignored at end of line %d\n",
		lineCount);
    } while (wordCount == 0);
    if (readType == PARSING) fprintf(stderr, "%4d:\t%s", lineCount, scriptLine);
    len = strlen(scriptTag);
    if ((scriptTag[len] == ':') && (wordCount > 1))
	printf("Warning: Extra characters ignored after label on line %d\n",
	    lineCount);
    return wordCount;
}


/************************************************************************
 *									*
 * The emitBytes() routine is used to write bytes to the output stream.	*
 * It is passed a pointer to the data to be written, along with the	*
 * length of the data field to write, and a flag indicating whether an	*
 * adjustment for "Endian"ism needs to be made.  If this flag is set to	*
 * NO_ENDIAN, or if the code is run on a LittleEndian machine, then	*
 * the bytes are written out directly.  If, however, the flag is set to	*
 * ENDIAN and this is NOT a LittleEndian machine, then the bytes are	*
 * written in reverse order.						*
 *									*
 ************************************************************************/

void emitBytes(char *value, int length, int adjustEndian)
{
    int i;
    byteCount += length;
    if (readType == SCANNING) return;
    if ((thisEndian == LITTLE_ENDIAN) || (adjustEndian == NO_ENDIAN))
	fwrite(value, length, 1, outFile);
    else if (adjustEndian == ENDIAN)
	for (i = length-1; i >= 0; i--) fwrite(value + i, 1, 1, outFile);
}


/************************************************************************
 *									*
 * The emitLength() routine is used to write a length specification to	*
 * the output stream.  This is the routine that handles implicit and	*
 * explicit VR specifications.  This handling has been placed here	*
 * since the way the length field is written changes depending on	*
 * whether implicit or explicit VR is being used.
 *									*
 ************************************************************************/

void emitLength(unsigned short group,
    VALUE_REPRESENTATION VR, UnsignedLong length)
{
    unsigned short shortLength = 0;

    /*
     * Explicit VR is used for this tag if the transfer syntax specifies
     * explicit_VR OR if the data is in group 0x0002 (File Meta Info).
     * THIS IS A MAJOR DICOM CHANGE, MADE 10/20/94.  FILE META INFO
     * WAS FORMERLY FORCED TO IMPLICIT VR; IT IS NOW FORCED TO EXPLICIT.
     * Explicit is not checked if the data is part of an item delimiter.
     */

    if (((explicit_VR == TRUE) || (group == 0x0002)) && (group != 0xfffe))
    {
	emitBytes(VR_label[VR], 2, NO_ENDIAN); /* Write explicit VR */
	/* Length for OB, OW, or SQ in explicit VR mode is UnsignedLong */
	if ((VR == VR_OB) || (VR == VR_OW) || (VR == VR_SQ))
	{
	    emitBytes((char *) &shortLength, 2, ENDIAN); /* 2 bytes padding */
	    emitBytes((char *) &length, 4, ENDIAN); /* 4 bytes length */
	}
	else /* If not OB, OW, or SQ, length is just unsigned short */
	{
	    shortLength = length;
	    emitBytes((char *) &shortLength, 2, ENDIAN);
	}
    }
    else emitBytes((char *) &length, 4, ENDIAN); /* Implicit VR mode */
}


/************************************************************************
 *									*
 * The emitUnsignedLong() routine handles UnsignedLong fields given	*
 * as either constants or symbolic references.  The syntax is somewhat	*
 * terse, but very powerful for specifying DICOM files and DICOMDIR.	*
 * A symbol preceded by '&' is replaced by the OFFSET of that symbols's	*
 * definition (location of label) from the beginning of the file.  For	*
 * example, &FIRST_PATIENT will be replaced with a value corresponding	*
 * to the offset of the FIRST_PATIENT: label from the beginning of the	*
 * file.  A symbol pair preceded by a '#' (with symbols separated by	*
 * ':' is replaced by the LENGTH of the bytes between the two symbols.	*
 * For example, #GROUP_START:GROUP_END is replaced with the number	*
 * of bytes between the labels GROUP_START: and GROUP_END:.  This is	*
 * very handy for specifying group lengths or lengths of sequences.	*
 *									*
 ************************************************************************/

void emitUnsignedLong(char *value)
{
    UnsignedLong longValue;		/* Value to be written */
    char *symbol1, *symbol2;		/* Symbols spanning and LENGTH */

    if (value[0] == '&') /* Look for offset specification */
    {
	/* Don't bother to look it up during the SCANNING pass */
	if (readType == SCANNING) longValue = 0L;
	else longValue = lookupSymbol(value+1);
    }
    else if (value[0] == '#') /* Look for length specification */
    {
	/* Don't bother to look it up during the SCANNING pass */
	if (readType == SCANNING) longValue = 0L;
	else /* Look up offset of each symbol, then subtract */
	{
	    symbol1 = strtok(value, ":");
	    symbol2 = strtok(NULL, ":");
	    longValue = lookupSymbol(symbol2) - lookupSymbol(symbol1+1);
	}
    }
    else /* Value is expressed as a constant */
    {
	if (strncmp(value, "0X", 2) == 0) /* Handle hex specification */
	    sscanf(value+2, "%lx", &longValue);
	else sscanf(value, "%lu", &longValue); /* Handle decimal spec */
    }
    emitBytes((char *) &longValue, 4, ENDIAN); /* Finally, write it out */
}


/************************************************************************
 *									*
 * The roundLength() function computes the length of a string rounded	*
 * up to the nearest even number of bytes.				*
 *									*
 ************************************************************************/

int roundLength(char *stringValue)
{
    int length;
    length = strlen(stringValue);
    if (((length / 2) * 2) == length) return length;
    return length + 1;
}


/************************************************************************
 *									*
 * The parseTag() function scans a DICOM tag in the form of (gggg,eeee)	*
 * and extracts the group and element fields.				*
 *									*
 ************************************************************************/
int parseTag(char *tagString, unsigned short *group, unsigned short *element)
{
    char *word;
    word = strtok(tagString, "(,");
    if (word == NULL) return -1;
    sscanf(word, "%hx", group);
    word = strtok(NULL, ")");
    if (word == NULL) return -1;
    sscanf(word, "%hx", element);
    return 0;
}

/************************************************************************
 *									*
 * The copyBytes() function copies bytes from an aux input file to the	*
 * destination file.							*
 *									*
 * NOTE: OBSCURE BUG!  It should be legal for the variable "auxBuffer"	*
 * to be automatic (local), rather than static, as it is not the	*
 * official buffer for the FILE *outFile.  It WOULD be improper to	*
 * allocate a file buffer as an automatic variable and then fail to	*
 * close the file before returning from the function where the buffer	*
 * was declared (the setvbuf routine would be used to cause that	*
 * variable to be used as I/O buffering).  Anyway, leaving auxBuffer	*
 * as automatic should be fine, and in fact, DOES run well using the	*
 * acc compiler on SunOS 4.1.3.  It causes improper output using the	*
 * Borland C 4.0 compiler under DOS.  So, the problem is avoided by	*
 * changing auxBuffer to "static".  Now THAT was a tough one to track	*
 * down!...								*
 *									*
 ************************************************************************/

void copyBytes(FILE *auxFile, UnsignedLong len)
{
    UnsignedLong bytesToRead, roundedLen;
    static unsigned char auxBuffer[BUFFER_LENGTH];
    const UnsignedLong Zero = 0; 

    bytesToRead = roundedLen = len;
    if ((len % 2) != 0) roundedLen++;
    if (readType == PARSING)
    {
	do {
	    if (bytesToRead > BUFFER_LENGTH)
	    {
		fread(auxBuffer, BUFFER_LENGTH, 1, auxFile);
		emitBytes((char *) auxBuffer, BUFFER_LENGTH, NO_ENDIAN);
		bytesToRead -= BUFFER_LENGTH;
	    }
	    else
	    {
		fread(auxBuffer, bytesToRead, 1, auxFile);
		emitBytes((char *) auxBuffer, bytesToRead, NO_ENDIAN);
		bytesToRead = 0;
	    }
	} while (bytesToRead != 0);

	if (len != roundedLen) emitBytes((char *) &Zero, 1, NO_ENDIAN);

    }
    else byteCount += roundedLen;
}

/************************************************************************
 *									*
 * The emitOB() function is called to output data with the "OB" VR.	*
 *									*
 ************************************************************************/

void emitOB(unsigned short group, unsigned short element, 
    VALUE_REPRESENTATION VR, char *value)
{
    FILE *auxFile;
    unsigned char oneByte;
    char errorMsg[80];
    char jpegFileName[MAX_NAME_SIZE], tableFileName[MAX_NAME_SIZE];
    int i, j, intByte;
    UnsignedLong len, roundedLen, longValue, addedOffset;
    UnsignedLong offsetCount, offsetTable[MAX_OFFSETS];
    const unsigned short ItemGroup=0xFFFE, ItemStart=0xE000, ItemEnd=0xE0DD;
    union{UnsignedLong asLong; unsigned char asByte[4];} swapValue;

    if (value[0] == '@') /* Named file has the data */
    {
	auxFile = fopen(value+1, "rb");
	if (auxFile == NULL)
	{
	    sprintf(errorMsg,"Unable to open auxiliary file %s",value+1);
	    badScript(errorMsg);
	}
	fseek(auxFile, 0, SEEK_END); /* seek to the end */
	roundedLen = len = ftell(auxFile); /* Gets file size */
	rewind(auxFile);
	if ((len % 2) != 0) roundedLen++;
	emitLength(group, VR, roundedLen);
	copyBytes(auxFile, len);
	if (fclose(auxFile)!=0) badScript("Error closing auxiliary input file");
    } /* if (value[0] == '@') */
    else if (value[0] == '*')
    {
	if ((group != 0x7FE0) || (element != 0x0010))
	    badScript("'*' only valid in (7FE0,0010)");
	len = 0xffffffff;
	emitLength(group, VR, len);

	strcpy(jpegFileName, scriptData+1);
	memset(tableFileName, 0, MAX_NAME_SIZE);
	strncpy(tableFileName, jpegFileName, strcspn(jpegFileName, "."));
	strcat(tableFileName, ".tab");

	auxFile = fopen(tableFileName, "rb");
	if (auxFile == NULL) badScript("Unable to open offset table file");
	fseek(auxFile, 0, SEEK_END); /* seek to the end */
	len = ftell(auxFile); /* Gets file size */
	rewind(auxFile);
	if (readType == PARSING)
	{
	    offsetCount = (len / 4);
	    if (offsetCount >= MAX_OFFSETS)
		badScript("Offset table too small: change #define MAX_OFFSETS");
	    for (i = 0; i < offsetCount; i++)
	    {
		if (thisEndian==LITTLE_ENDIAN)
		    fread(&offsetTable[i], 4, 1, auxFile);
		else
		{
		    for (j = 3; j >= 0; j--)
		        fread(&swapValue.asByte[j], 1, 1, auxFile);
		    offsetTable[i] = swapValue.asLong;
		}
	    }
	} /* if (readType == PARSING) */
	if (fclose(auxFile) != 0) badScript("Error closing offset table file");
	/* Write item tag, length of offset table */
	if (readType == PARSING) 
	    printf("\tCreating offset table (%04X,%04X) <<%lu bytes>>\n",
		ItemGroup, ItemStart, len);
	emitBytes((char *) &ItemGroup, 2, ENDIAN);
	emitBytes((char *) &ItemStart, 2, ENDIAN);
	emitBytes((char *) &len, 4, ENDIAN);

	auxFile = fopen(jpegFileName, "rb");
	if (auxFile == NULL) badScript("Unable to open pixel file");
	fseek(auxFile, 0, SEEK_END); /* seek to the end */
	len = ftell(auxFile); /* Gets file size */
	rewind(auxFile);
	offsetTable[offsetCount] = len;

	addedOffset = 0;
	for (i = 0; i < offsetCount; i++)
	{
	    if (((offsetTable[i]+addedOffset) % 2) != 0) addedOffset++;
	    longValue = offsetTable[i] + addedOffset;
	    emitBytes((char *) &longValue, 4, ENDIAN);
	    addedOffset += 8;
	}

	for (i = 0; i < offsetCount; i++)
	{
	    roundedLen = len = offsetTable[i+1] - offsetTable[i];
	    if ((len % 2) != 0) roundedLen++;
	    if (readType == PARSING)
		printf("\tEncapsulating frame %d <<%lu bytes>>\n",
		    i+1, roundedLen);
	    emitBytes((char *) &ItemGroup, 2, ENDIAN);
	    emitBytes((char *) &ItemStart, 2, ENDIAN);
	    emitBytes((char *) &roundedLen, 4, ENDIAN);
	    copyBytes(auxFile, len);
	}
	if (readType == PARSING) 
	    printf("\tEnd of encapsulated pixels (%04X,%04X)\n",
		ItemGroup, ItemEnd);
	len = 0;
	emitBytes((char *) &ItemGroup, 2, ENDIAN);
	emitBytes((char *) &ItemEnd, 2, ENDIAN);
	emitBytes((char *) &len, 4, ENDIAN);
    } /* else if (value[0] == '*') */
    else
    {
	if ((strncmp(value,"0X",2) != 0) || (strlen(value)<=2))
	    badScript("Improperly formed hex constant");
	len = roundLength(value);
	if (len != strlen(value))
	    badScript("Uneven byte count in hex constant");
	len = (len - 2) / 2;
	emitLength(group, VR, len);
	for (i = 0; i < (len*2); i += 2)
	{
	    sscanf(value+i+2, "%2x", &intByte);
	    oneByte = intByte;
	    emitBytes((char *) &oneByte, 1,NO_ENDIAN);
	}
    }
}

/************************************************************************
 *									*
 * The dicomMsg() routine outputs DICOM messages based on the group and	*
 * element values it is passed.  The data field for the messages is	*
 * passed as a character string, and is converted as appropriate based	*
 * on the VR that corresponds to the tag (group and element).		*
 *									*
 ************************************************************************/

void dicomMsg(unsigned short group, unsigned short element, char *value)
{
    VALUE_REPRESENTATION VR;
    struct dictStruct *dict;
    UnsignedLong len, longValue;
    unsigned short shortValue, tagGroup, tagElement;
    unsigned short VM = 1;
    char errorMsg[80];
    const char Space = ' ';
    int i;
    char *oneValue;

    if ((dict = dictLookup(group, element)) == (struct dictStruct *) NULL)
    {
	sprintf(errorMsg, "(%04X,%04X) not in dictionary\n", group, element);
	badScript(errorMsg);
    }
    VR = dict->VR;
    emitBytes((char *) &group, 2, ENDIAN);
    emitBytes((char *) &element, 2, ENDIAN);

    for (i = 0; i < strlen(value); i++) if (value[i] == '\\') VM++;
    if (strlen(value) == 0) emitLength(group, VR, 0);
    else switch(VR)
    {
	case VR_AT: /* Attribute Tag */
		    parseTag(value, &tagGroup, &tagElement);
		    emitLength(group, VR, 4);
		    emitBytes((char *) &tagGroup, 2, ENDIAN);
		    emitBytes((char *) &tagElement, 2, ENDIAN);
		    break;
	case VR_SQ: /* Sequence */
		    if (strncmp(value, "0X", 2) == 0)
			sscanf(value+2, "%lX", &longValue);
		    else sscanf(value, "%lu", &longValue);
		    emitLength(group, VR, longValue);
		    break;
	case VR_DATA: /* Data Delimiter */
		    emitUnsignedLong(value);
		    break;
	case VR_UL: /* Unsigned Long */
		    emitLength(group, VR, 4 * VM);
		    oneValue = strtok(value, "\\");
		    while (oneValue != NULL)
		    {
			emitUnsignedLong(oneValue);
			oneValue = strtok(NULL, "\\");
		    }
		    break;
	case VR_US: /* Unsigned Short */
		    emitLength(group, VR, 2 * VM);
		    oneValue = strtok(value, "\\");
		    while (oneValue != NULL)
		    {
			if (strncmp(oneValue, "0X", 2) == 0)
			    sscanf(oneValue+2, "%hx", &shortValue);
			else sscanf(oneValue, "%hd", &shortValue);
			emitBytes((char *) &shortValue, 2, ENDIAN);
			oneValue = strtok(NULL, "\\"); 
		    }
		    break;
	case VR_UI: /* Unique Identifier */
		    /*
		     * Check for any transfer syntax other than default.
		     * If present, then switch to explicit VR mode (see
		     * Part 5, sections 10.2 and A.4)
		     */
		    if ((group == 0x0002) && (element == 0x0010)
			&& (strcmp(value, LITTLE_ENDIAN_IMPLICIT_UID) != 0))
		    {
			explicit_VR = TRUE;
			if (strcmp(value, LOSSLESS_JPEG_14_UID) == 0)
			    encodedPixels = TRUE;
		    } /* case falls thru to VR_ST */
	case VR_CS: /* Control String */
	case VR_DA: /* Date */
	case VR_DS: /* Decimal String */
	case VR_IS: /* Integer String */
	case VR_LO: /* Long String */
	case VR_PN: /* Person Name */
	case VR_SH: /* Short String */
	case VR_TM: /* Time */
	case VR_LT: /* Long Text */
	case VR_ST: /* Short Text */
		    len = roundLength(value);
		    emitLength(group, VR, len);
		    if (VR == VR_UI) emitBytes(value, len, NO_ENDIAN);
		    else
		    {
		        emitBytes(value, strlen(value), NO_ENDIAN);
		        if (len != strlen(value))
			    emitBytes((char *) &Space, 1, NO_ENDIAN);
		    }
		    break;
	case VR_OW: /* Other word field */
	case VR_OB: /* Other byte field */
		    emitOB(group, element, VR, value);
		    break;
	case VR_UNKNOWN: /* Undefined -- possible retired */
	default:    sprintf(errorMsg, "Unknown VR type: %s", VR_label[VR]);
		    badScript(errorMsg);
    }
}


/************************************************************************
 *									*
 * The buildFile() routine creates a DICOM File.  The output file is	*
 * opened, the necessary File Meta Information is created, and the	*
 * specified DICOM messages that comprise the file are written.		*
 *									*
 ************************************************************************/

void buildFile(void)
{
    char preamble[128];
    char msg[80];
    unsigned short group, element;
    long fileSpot;
    int lineSpot, wordCount, i;

    /* Open file for output */
    if ((outFile=fopen(scriptData, "wb")) == NULL)
    {
	sprintf(msg, "Unable to open output file \"%s\"", scriptData);
	badScript(msg);
    }
    outFileOpen = TRUE;
    byteCount = 0L;
    symbolCount = 0;

    memset(preamble, '\0', 128);
    emitBytes(preamble, 128, NO_ENDIAN);
    emitBytes("DICM", 4, NO_ENDIAN);

    fileSpot = ftell(scriptFile); /* Remember where we are in the file */
    lineSpot = lineCount; /* and the line we were on */
    readType = SCANNING;
    while (((wordCount=readScript()) != 0) && (strcmp(scriptTag, "END")!=0))
    {
	if (scriptTag[strlen(scriptTag)-1] == ':')
	{
	    scriptTag[strlen(scriptTag)-1] = '\0';
	    /* check for duplicate symbol */
	    if (symbolCount > 0) for (i=0; i<symbolCount; i++)
	    {
		if (strcmp(symbolTable[i].name, scriptTag)==0)
		    badScript("Duplicate label");
	    }
	    strcpy(symbolTable[symbolCount].name, scriptTag);
	    symbolTable[symbolCount++].value = byteCount;
	    /*printf("Value of symbol %s is %lX\n",scriptTag,byteCount);*/
	    if (symbolCount == MAX_SYMBOLS)
		badScript("Too many symbols: change #define MAX_SYMBOLS");
	}
	else
	{
	    if (parseTag(scriptTag, &group, &element) == -1)
	    {
		sprintf(msg, "Improperly formed tag: %s", scriptTag);
		badScript(msg);
	    }
	    if (wordCount > 1) dicomMsg(group, element, scriptData);
	    else dicomMsg(group, element, "");
	}
    }
    /* Return to script file location where we started symbol table pass */
    fseek(scriptFile, fileSpot, 0);
    lineCount = lineSpot;
    readType = PARSING;

    while ((wordCount = readScript()) != 0)
    {
	if (strcmp(scriptTag, "END") == 0)
	{
	    if (fclose(outFile) != 0) badScript("Error closing output file");
	    outFileOpen = FALSE;
	    return;
	}
	if (scriptTag[strlen(scriptTag)-1] != ':')
	{
	    if (parseTag(scriptTag, &group, &element) == -1)
		badScript("Improperly formed tag");
	    if (wordCount > 1) dicomMsg(group, element, scriptData);
	    else dicomMsg(group, element, "");
	}
    }
    badScript("Unexpected End-Of-File");
}


/************************************************************************
 *									*
 * The main() part of the program does little except to call the	*
 * necessary subroutines to read, parse, and output the data.		*
 *									*
 ************************************************************************/

int main(int argc, char **argv)
{
    int wordCount;

    fprintf(stderr, "------------------------------------");
    fprintf(stderr, "------------------------------------\n");
    fprintf(stderr, "%s\n", COPYRIGHT1);
    fprintf(stderr, "%s\n", COPYRIGHT2);
    fprintf(stderr, "%s\n", COPYRIGHT3);
    fprintf(stderr, "------------------------------------");
    fprintf(stderr, "------------------------------------\n\n");

    thisEndian = checkEndian();
    if (argc != 2) /* Be sure there is a name given */
    {
	fprintf(stderr, "No script file specified\n");
	exit(EXIT_FAILURE);
    }
    if ((scriptFile=fopen(argv[1], "rt")) == NULL) /* Valid file? */
    {
	fprintf(stderr, "Unable to open script file \"%s\"\n",
	    argv[1]);
	exit(EXIT_FAILURE);
    }

    scriptFileOpen = TRUE;
    readType = PARSING;
    while ((wordCount = readScript()) != 0)
    {
	if ((strcmp(scriptTag, "BEGIN") != 0) || (wordCount != 2))
	    badScript("BEGIN expected");
	buildFile();
    }
    fclose(scriptFile);
    return(0);
}


