/*************************************************************************
 *************************************************************************
 **									**
 **	 		Copyright (c) 1994, 1995 			**
 **	    Brown University Institute for Medical Computing		**
 **	      and the American College of Cardiology (ACC)		**
 **									**
 ** The software and support documentation for the American College of	**
 ** Cardiology's 1995 demonstration of DICOM interchange media were	**
 ** developed at the Brown University Institute for Medical Computing.	**
 **									** 
 ** THIS SOFTWARE IS MADE AVAILABLE "AS IS", AND NEITHER THE AMERICAN	**
 ** COLLEGE OF CARDIOLOGY (ACC) NOR BROWN UNIVERSITY MAKE ANY WARRANTY	**
 ** ABOUT THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR FITNESS	**
 ** FOR ANY PARTICULAR USE, FREEDOM FROM DEFECTS, OR ITS CONFORMANCE TO	**
 ** ANY SPECIFICATION.  THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE	**
 ** OF THE SOFTWARE IS WITH THE USER.					**
 **									** 
 ** Copyright of the software and supporting documentation is jointly	**
 ** owned by the ACC and Brown University, and free access is hereby	**
 ** granted as a license to use this software, copy this software and	**
 ** prepare derivative works based upon this software.  However, any	**
 ** distribution of the source code or executable versions of this	**
 ** software, or of the	 supporting documentation (or of derivative	**
 ** works of the software or supporting documentation) must include	**
 ** this copyright notice and explanation.				**
 **									** 
 ** Author: Jonathan L. Elion MD, FACC					**
 **									**
 ************************************************************************* 
 *************************************************************************/ 

/************************************************************************
 *									*
 * This utility reads a DICOM file file, parses the fields, and prints	*
 * out the information in human-readable format.			*
 *									*
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "acc95.h"


/************************************************************************
 * The following items are GLOBAL, and are used by several routines.	*
 ************************************************************************/

int	explicit_VR = FALSE,		/* Implicit versus Explicit VR */
	encodedPixels = FALSE,		/* Are pixels compressed? */
	sourceOut = FALSE;		/* Should output be a source file */
FILE	*inFile;			/* Input file */
int	thisEndian;
UnsignedLong offset=0;
char *VR_label[] =
   {"AE", "AS", "AT", "CS", "DA", "DS", "DT", "FL", "FD", "IS", "LO",
    "LT", "OB", "OW", "PN", "SH", "SL", "SQ", "SS", "ST", "TM", "UI",
    "UL", "US", "**", "??"};

/************************************************************************
 *									*
 * The readBytes() routine is used to read bytes from the input stream.	*
 * It is passed a pointer to the buffer to read into, along with the	*
 * length of the data field to read, and a flag indicating whether an	*
 * adjustment for "Endian"ism needs to be made.  If this flag is set to	*
 * NO_ENDIAN, or if the code is run on a LittleEndian machine, then	*
 * the bytes are read in directly.  If, however, the flag is set to	*
 * ENDIAN and this is NOT a LittleEndian machine, then the byte order	*
 * is reversed.								*
 *									*
 ************************************************************************/

int readBytes(char *value, int length, int adjustEndian)
{
    int i;
    offset += length;
    if ((thisEndian == LITTLE_ENDIAN) || (adjustEndian == NO_ENDIAN))
	return fread(value, length, 1, inFile);
    else if (adjustEndian == ENDIAN) for (i = length-1; i >= 0; i--)
    {
	if (fread(value + i, 1, 1, inFile) != 1) return 0;
    }
    return 1;
}


/************************************************************************
 *									*
 * The badFile() function reports an error message and exits.		*
 *									*
 ************************************************************************/

void badFile(char *msg)
{
    printf("%s\n", msg);
    exit(EXIT_FAILURE);
}


/************************************************************************
 *									*
 * The copyBytes() function copies bytes to an aux output file.		*
 *									*
 ************************************************************************/

void copyBytes(char *fileName, int fileNumber, UnsignedLong len)
{
    FILE *auxFile;
    char auxFileName[128];
    UnsignedLong bytesToRead;
    static unsigned char auxBuffer[BUFFER_LENGTH];

    sprintf(auxFileName, "%s.%03d", fileName, fileNumber);
    if ((auxFile=fopen(auxFileName, "wb")) == NULL) /* Open file */
    {
	printf("Unable to open output file \"%s\"\n", auxFileName);
	exit(EXIT_FAILURE);
    }

    bytesToRead = len;
    do {
	if (bytesToRead > BUFFER_LENGTH)
	{
	    readBytes((char *) auxBuffer, BUFFER_LENGTH, NO_ENDIAN);
	    fwrite(auxBuffer, BUFFER_LENGTH, 1, auxFile);
	    bytesToRead -= BUFFER_LENGTH;
	}
	else
	{
	    readBytes((char *) auxBuffer, bytesToRead, NO_ENDIAN);
	    fwrite(auxBuffer, bytesToRead, 1, auxFile);
	    bytesToRead = 0;
	}
    } while (bytesToRead != 0);
    if (fclose(auxFile) != 0) badFile("Error closing file aux output file");
    printf("@%s", auxFileName);
}


/************************************************************************
 *									*
 * The main() routine							*
 *									*
 ************************************************************************/

int main(int argc, char **argv)
{
    unsigned char preamble[128];
    char buffer[1024];
    char outFileName[128];
    char VR_Tag[] = "??", commentDelim[] = " ";
    unsigned short group, lastGroup, element, tagGroup, tagElement, shortValue;
    UnsignedLong length, longValue;
    struct dictStruct *dict;
    char *element_description;
    VALUE_REPRESENTATION VR;
    int i, keepLooping, doingGroupLength;
    int outFileNumber = 0;

    thisEndian = checkEndian();
    if (argc < 2) /* Be sure there is a name given */
    {
	printf("No input file specified\n");
	exit(EXIT_FAILURE);
    }
    if ((inFile=fopen(argv[1], "rb")) == NULL) /* Valid file? */
    {
	printf("Unable to open input file \"%s\"\n", argv[1]);
	exit(EXIT_FAILURE);
    }
    if (argc > 2)
    {
	commentDelim[0] = ';';
	sourceOut = TRUE;
	strcpy(outFileName, argv[2]);
    }
    else commentDelim[0] = '\0';

    printf("%s------------------------------------", commentDelim);
    printf("------------------------------------\n");
    printf("%s%s\n", commentDelim, COPYRIGHT1);
    printf("%s%s\n", commentDelim, COPYRIGHT2);
    printf("%s%s\n", commentDelim, COPYRIGHT3);
    if (sourceOut == TRUE)
    {
	printf(";\n;This file was generated automatically from ");
	printf("file \"%s\"\n;\n", argv[1]);
    }
    printf("%s------------------------------------", commentDelim);
    printf("------------------------------------\n\n");

    if (sourceOut == TRUE) printf("BEGIN %s\n\n", outFileName);

    if (readBytes((char *) preamble, 128, NO_ENDIAN) != 1)
	badFile("Unexpected EOF");
    if (readBytes((char *) preamble, 4, NO_ENDIAN) != 1)
	badFile("Unexpected EOF");
    if (strncmp((char *) preamble, "DICM", 4) != 0)
    {
	printf("Error: Not a valid DICOM file.  ");
	printf("Will try to read as non-Part 10 file.\n");
	rewind(inFile);
    }
    keepLooping = readBytes((char *) &group, 2, ENDIAN);
    lastGroup = group;
    while(keepLooping == 1)
    {
	if (lastGroup != group)
	{
	    if ((sourceOut == TRUE) && (doingGroupLength == 1))
	    {
		printf("GROUP%X_END:\n", lastGroup);
		doingGroupLength = 0;
	    }
	    printf("\n");
	    lastGroup = group;
	}
	if (sourceOut == FALSE) printf("%08lX:  ", offset-2);
	if (readBytes((char *) &element, 2, ENDIAN) != 1)
	    badFile("Unexpected EOF");
	dict = dictLookup(group, element);
	if (dict == (struct dictStruct *) NULL)
	{
	    sprintf(buffer, "Undefined tag: (%04X,%04X)", group, element);
	    /* An unknown element can be handled only if in Explicit mode */
	    if (explicit_VR == TRUE)
	    {
		printf("%s\n           ", buffer);
		VR = VR_UNKNOWN;
		element_description = "*** Not in dictionary! ***";
	    }
	    else badFile(buffer);
	}
	else
	{
	    VR = dict->VR;
	    element_description = dict->description;
	}
	if (((explicit_VR==TRUE)||(group == 0x0002)) && (group != 0xfffe))
	{
	    readBytes(VR_Tag, 2, NO_ENDIAN);
	    /* If element not in dictionary, use specified (explicit) VR */
	    if (VR == VR_UNKNOWN)
	    {
		/* Convert VR as specified to corresponding enum type */
		for (i = (int) VR_AE; i <= (int) VR_DATA; i++)
		{
		    if (strcmp(VR_label[i], VR_Tag) == 0)
		    {
			VR = (VALUE_REPRESENTATION) i;
			break;
		    }
		}
		if (VR == VR_UNKNOWN)
		{
	    	    sprintf(buffer, "Explicit VR %s is not defined", VR_Tag);
		    badFile(buffer);
		}
	    }
	    if ((VR != VR_OB) && (strcmp(VR_Tag, "OW") != 0) &&
		(group != 0x7FE0) && (element != 0x0010) &&
		(strcmp(VR_Tag, VR_label[VR]) != 0))
	    {
		printf("%sWarning: (%04X,%04X) Explicit VR tag %s ",
		    commentDelim, group, element, VR_Tag);
		printf("does not match dictionary VR tag %s\n", VR_label[VR]);
	    }
	    if ((VR == VR_OB) || (VR == VR_OW) || (VR == VR_SQ))
	    {
		readBytes((char *) &shortValue, 2, ENDIAN);
		readBytes((char *) &length, 4, ENDIAN);
	    }
	    else
	    {
		readBytes((char *) &shortValue, 2, ENDIAN);
		length = shortValue; 
	    }
	}
	else readBytes((char *) &length,  4, ENDIAN);

	if (sourceOut == FALSE) printf("(%04X,%04X) [%s/%lu]\t",
	    group, element, VR_label[VR], length);
	else
	{
	    printf("    (%04X,%04X) ", group, element);
	}

	switch(VR)
	{
	    case VR_AT: /* Attribute Tag */
		readBytes((char *) &tagGroup, 2, ENDIAN);
		readBytes((char *) &tagElement, 2, ENDIAN);
		printf("(%04X,%04X)", tagGroup, tagElement);
		break;
	    case VR_SQ: /* Sequence */
		printf("0x%08lX", length);
		break;
	    /* case VR_FL: Floating Point Single 
	    /* case VR_FD: Floating Point Double
		break; */
	    case VR_SL: /* Signed Long */
		if (length == 4)
		{
		    readBytes((char *) &longValue, 4, ENDIAN);
		    if (sourceOut == TRUE) printf("%l", longValue);
		    else printf("%l (0x%04lX)", longValue, longValue);
		}
		else for (i = 0; i < (length/4); i++)
		{
		    readBytes((char *) longValue, 4, ENDIAN);
		    if (i > 0) printf("\\");
		    printf("%l", longValue);
		}
		break;
	    case VR_UL: /* Unsigned Long */
		if (length == 4)
		{
		    readBytes((char *) &longValue, 4, ENDIAN);
		    if (sourceOut == TRUE)
		    {
			if (element == 0)
			{
			    printf("#GROUP%X_START:GROUP%X_END",
			    	group, group);
			    doingGroupLength = 1;
			}
			 else printf("%lu", longValue);
		    }
		    else printf("%lu (0x%04lX)", longValue, longValue);
		}
		else for (i = 0; i < (length/4); i++)
		{
		    readBytes((char *) longValue, 4, ENDIAN);
		    if (i > 0) printf("\\");
		    printf("%lu", longValue);
		}
		break;
	    case VR_SS: /* Signed Short */
		if (length == 2)
		{
		    readBytes((char *) &shortValue, 2, ENDIAN);
		    if (sourceOut == TRUE) printf("%d", shortValue);
		    else printf("%d (0x%04X)", shortValue, shortValue);
		}
		else for (i = 0; i < (length/2); i++)
		{
		    readBytes((char *) &shortValue, 2, ENDIAN);
		    if (i > 0) printf("\\");
		    printf("%d", shortValue);
		}
		break;
	    case VR_US: /* Unsigned Short */
		if (length == 2)
		{
		    readBytes((char *) &shortValue, 2, ENDIAN);
		    if (sourceOut == TRUE) printf("%u", shortValue);
		    else printf("%u (0x%04X)", shortValue, shortValue);
		}
		else for (i = 0; i < (length/2); i++)
		{
		    readBytes((char *) &shortValue, 2, ENDIAN);
		    if (i > 0) printf("\\");
		    printf("%u", shortValue);
		}
		break;
	    case VR_AE: /* Application Entity */
	    case VR_AS: /* Age String */
	    case VR_CS: /* Control String */
	    case VR_DA: /* Date */
	    case VR_DS: /* Decimal String */
	    case VR_DT: /* Date Time */
	    case VR_IS: /* Integer String */
	    case VR_LO: /* Long String */
	    case VR_PN: /* Person Name */
	    case VR_SH: /* Short String */
	    case VR_TM: /* Time */
	    case VR_ST: /* Short Text */
	    case VR_LT: /* Long Text */
	    case VR_UI: /* Unique Identifier */
		memset(buffer, '\0', 1024);
		readBytes(buffer, length, NO_ENDIAN);
		printf("\"%s\"", buffer);
		if ((group == 0x0002) && (element == 0x0010) &&
		    (strcmp(buffer, LITTLE_ENDIAN_IMPLICIT_UID) !=0))
		{
		    explicit_VR = TRUE;
		    /* printf("*** Switching to explicit VR mode ***\n");*/
		    if (strcmp(buffer, LOSSLESS_JPEG_14_UID) == 0)
			encodedPixels = TRUE;
		}
		break;
	    case VR_UNKNOWN: /* Possibly a retired field */
	    case VR_DATA: /* Data tag */
	    case VR_OB: /* Other Byte */
		if ((length == 0) || (length == 0xFFFFFFFF)) break;
		if (length <= 16)
		{
		    printf("0x");
		    for (i=0; i<length; i++)
		    {
			readBytes(buffer, 1, NO_ENDIAN);
			printf("%02X", buffer[0]);
		    }
		}
		else if (sourceOut == TRUE)
		{
		    copyBytes(outFileName, outFileNumber, length);
		    outFileNumber++;
		}
		else
		{
		    printf("<< %lu bytes of data skipped >>", length);
		    fseek(inFile, length, SEEK_CUR);
		    offset += length;
		}
		break;
	    case VR_OW: /* Other Word */
		if (length <= 16)
		{
		    printf("0x");
		    for (i=0; i<length; i++)
		    {
			readBytes(buffer, 2, ENDIAN);
			printf("%02X%02X", buffer[0], buffer[1]);
		    }
		}
		else if (sourceOut == TRUE)
		{
		    copyBytes(outFileName, outFileNumber, length);
		    outFileNumber++;
		}
		else
		{
		    printf("<< %lu bytes of data skipped >>", length);
		    fseek(inFile, length, SEEK_CUR);
		    offset += length;
		}
		break;
	    default:
		sprintf(buffer, "Can''t handle VR type: %s", VR_label[VR]);
		badFile(buffer);
	}
	printf("\t; %s\n", element_description);
	if ((sourceOut == TRUE) && (element == 0))
	    printf("GROUP%X_START:\n", group);
	keepLooping = readBytes((char *) &group,   2, ENDIAN);
    }
    fclose(inFile);
    if (sourceOut == TRUE)
    {
	if (doingGroupLength == 1) printf("GROUP%X_END:\n", lastGroup);
	printf("END\n");
    }
    return(0);
}

